"""Partial implementation of the deprecated audioop module.

Only supports:
  - widths 1, 2, and 4
  - signed samples
  - bias, tomono, tostereo, lin2lin, ratecv
"""

import math
import struct
from typing import Final, List, Literal, Optional, Tuple, Union

BufferType = Union[bytes, bytearray]
Endian = Literal["big", "little"]
State = Tuple[int, Tuple[Tuple[int, ...], ...]]

# width = (_, 1, 2, _, 4)
_MAX_VALS: Final = [0, 0x7F, 0x7FFF, 0, 0x7FFFFFFF]
_MIN_VALS: Final = [0, -0x80, -0x8000, 0, -0x80000000]
_SIGNED_FORMATS: Final = ["", "b", "h", "", "i"]
_UNSIGNED_FORMATS: Final = ["", "B", "H", "", "I"]
_ENDIAN_FORMAT_PREFIX: Final = {"big": ">", "little": "<"}


def check_size(size: int) -> None:
    if size not in (1, 2, 4):
        raise ValueError(f"Size should be 1, 2, 4. Got {size}")


def check_parameters(fragment_length: int, size: int) -> None:
    check_size(size)
    if (fragment_length % size) != 0:
        raise ValueError(
            "Not a whole number of frames: "
            f"fragment_length={fragment_length}, size={size}"
        )


def _get_struct_format(width: int, endian: Endian) -> str:
    struct_format = _SIGNED_FORMATS[width]
    if width == 1:
        return struct_format

    try:
        return _ENDIAN_FORMAT_PREFIX[endian] + struct_format
    except KeyError as err:
        raise ValueError(f"Endian should be 'big' or 'little'. Got {endian}") from err


def fbound(val: float, min_val: float, max_val: float) -> int:
    if val > max_val:
        val = max_val
    elif val < (min_val + 1):
        val = min_val

    val = math.floor(val)

    return int(val)


def tomono(
    fragment: BufferType, width: int, lfactor: float, rfactor: float
) -> BufferType:
    fragment_length = len(fragment)
    check_parameters(fragment_length, width)

    max_val = _MAX_VALS[width]
    min_val = _MIN_VALS[width]
    struct_format = _SIGNED_FORMATS[width]
    result = bytearray(fragment_length // 2)

    for i in range(0, fragment_length, width * 2):
        val_left = struct.unpack_from(struct_format, fragment, i)[0]
        val_right = struct.unpack_from(struct_format, fragment, i + width)[0]
        val_mono = (val_left * lfactor) + (val_right * rfactor)
        sample_mono = fbound(val_mono, min_val, max_val)
        struct.pack_into(struct_format, result, i // 2, sample_mono)

    return result


def tostereo(
    fragment: BufferType, width: int, lfactor: float, rfactor: float
) -> BufferType:
    fragment_length = len(fragment)
    check_parameters(fragment_length, width)

    max_val = _MAX_VALS[width]
    min_val = _MIN_VALS[width]
    struct_format = _SIGNED_FORMATS[width]
    result = bytearray(fragment_length * 2)

    for i in range(0, fragment_length, width):
        val_mono = struct.unpack_from(struct_format, fragment, i)[0]
        sample_left = fbound(val_mono * lfactor, min_val, max_val)
        sample_right = fbound(val_mono * rfactor, min_val, max_val)
        struct.pack_into(struct_format, result, i * 2, sample_left)
        struct.pack_into(struct_format, result, (i * 2) + width, sample_right)

    return result


def _get_sample32(fragment: BufferType, width: int, index: int, endian: Endian) -> int:
    """Extract a sample and convert it to 32-bit representation."""
    if width == 1:
        return struct.unpack_from("b", fragment, index)[0] << 24

    if width == 2:
        return (
            struct.unpack_from(_get_struct_format(width, endian), fragment, index)[0]
            << 16
        )

    if width == 4:
        return struct.unpack_from(_get_struct_format(width, endian), fragment, index)[0]

    raise ValueError(f"Invalid width: {width}")


def _set_sample32(
    fragment: bytearray, width: int, index: int, sample: int, endian: Endian
) -> None:
    """Set a sample from 32-bit representation."""
    if width == 1:
        struct.pack_into("b", fragment, index, sample >> 24)
    elif width == 2:
        struct.pack_into(
            _get_struct_format(width, endian), fragment, index, sample >> 16
        )
    elif width == 4:
        struct.pack_into(_get_struct_format(width, endian), fragment, index, sample)
    else:
        raise ValueError(f"Invalid width: {width}")


def bias(
    fragment: BufferType, width: int, bias: int, endian: Endian = "little"
) -> BufferType:
    fragment_length = len(fragment)
    check_parameters(fragment_length, width)

    result = bytearray(fragment_length)
    mask = (1 << (8 * width)) - 1

    for i in range(0, fragment_length, width):
        # Add bias with wrap-around on overflow, matching audioop.bias.
        sample = int.from_bytes(fragment[i : i + width], endian)
        sample = (sample + bias) & mask
        result[i : i + width] = sample.to_bytes(width, endian)

    return result


def lin2lin(
    fragment: BufferType, width: int, new_width: int, endian: Endian = "little"
) -> BufferType:
    if width == new_width:
        return fragment

    fragment_length = len(fragment)
    check_parameters(fragment_length, width)
    check_size(new_width)

    result = bytearray(int((fragment_length / width) * new_width))

    j = 0
    for i in range(0, fragment_length, width):
        sample = _get_sample32(fragment, width, i, endian)
        _set_sample32(result, new_width, j, sample, endian)
        j += new_width

    return result


def ratecv(
    fragment: BufferType,
    width: int,
    nchannels: int,
    inrate: int,
    outrate: int,
    state: Optional[State],
    weightA: int = 1,
    weightB: int = 0,
) -> Tuple[bytearray, Optional[State]]:
    fragment_length = len(fragment)
    check_size(width)
    if nchannels < 1:
        raise ValueError(f"Number of channels should be >= 1, got {nchannels}")
    bytes_per_frame = width * nchannels
    if (weightA < 1) or (weightB) < 0:
        raise ValueError(
            "weightA should be >= 1, weightB should be >= 0, "
            f"got weightA={weightA}, weightB={weightB}"
        )

    if (fragment_length % bytes_per_frame) != 0:
        raise ValueError("Not a whole number of frames")

    if (inrate <= 0) or (outrate <= 0):
        raise ValueError("Sampling rate not > 0")

    d = math.gcd(inrate, outrate)
    inrate //= d
    outrate //= d

    d = math.gcd(weightA, weightB)
    weightA //= d
    weightB //= d

    prev_i: List[int] = [0] * nchannels
    cur_i: List[int] = [0] * nchannels

    if state is None:
        d = -outrate
        # prev_i and cur_i are already zeroed
    else:
        d, samps = state
        if len(samps) != nchannels:
            raise ValueError("Illegal state argument")

        for chan_index, channel in enumerate(samps):
            prev_i[chan_index], cur_i[chan_index] = channel

    input_frames = fragment_length // bytes_per_frame
    output_frames = int(math.ceil(input_frames * (outrate / inrate)))

    # Approximate version used in C code to avoid overflow:
    # q = 1 + ((input_frames - 1) // inrate)
    # output_frames = q * outrate * bytes_per_frame

    result = bytearray(output_frames * bytes_per_frame)
    struct_format = _SIGNED_FORMATS[width]

    input_index = 0
    output_index = 0
    while True:
        while d < 0:
            if input_frames == 0:
                samps = tuple(
                    (prev_i[chan], cur_i[chan]) for chan in range(0, nchannels)
                )

                # NOTE: It's critical that result is clipped here
                return result[:output_index], (d, samps)

            for chan in range(0, nchannels):
                prev_i[chan] = cur_i[chan]
                cur_i[chan] = struct.unpack_from(struct_format, fragment, input_index)[
                    0
                ]
                input_index += width
                cur_i[chan] = ((weightA * cur_i[chan]) + (weightB * prev_i[chan])) // (
                    weightA + weightB
                )

            input_frames -= 1
            d += outrate
        while d >= 0:
            for chan in range(0, nchannels):
                sample = int(
                    (
                        (float(prev_i[chan]) * float(d))
                        + (float(cur_i[chan]) * (float(outrate) - float(d)))
                    )
                    / float(outrate)
                )
                struct.pack_into(struct_format, result, output_index, sample)
                output_index += width
            d -= inrate

    return result, None



"""Runs mDNS zeroconf service for Home Assistant discovery."""

import logging
import socket
import uuid
from typing import Optional

_LOGGER = logging.getLogger(__name__)

try:
    from zeroconf.asyncio import AsyncServiceInfo, AsyncZeroconf
except ImportError:
    _LOGGER.fatal("pip install zeroconf")
    raise

MDNS_TARGET_IP = "224.0.0.251"


class HomeAssistantZeroconf:
    """ZeroConf (mDNS) discovery for Home Assistant."""

    def __init__(
        self, port: int, name: Optional[str] = None, host: Optional[str] = None
    ) -> None:
        self.port = port
        self.name = name or _get_mac_address()

        if not host:
            test_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            test_sock.setblocking(False)
            test_sock.connect((MDNS_TARGET_IP, 1))
            host = test_sock.getsockname()[0]
            _LOGGER.debug("Detected IP: %s", host)

        assert host
        self.host = host
        self._aiozc = AsyncZeroconf()

    async def register_server(self) -> None:
        """Register ZeroConf server."""

        service_info = AsyncServiceInfo(
            "_wyoming._tcp.local.",
            f"{self.name}._wyoming._tcp.local.",
            addresses=[socket.inet_aton(self.host)],
            port=self.port,
        )
        await self._aiozc.async_register_service(service_info)
        _LOGGER.debug("Zeroconf discovery enabled: %s", service_info)


def _get_mac_address() -> str:
    """Return MAC address formatted as hex with no colons."""
    return "".join(
        # pylint: disable=consider-using-f-string
        ["{:02x}".format((uuid.getnode() >> ele) & 0xFF) for ele in range(0, 8 * 6, 8)][
            ::-1
        ]
    )



"""Server tests."""

import asyncio
import os
import signal
import socket
import tempfile
from pathlib import Path

import pytest

from wyoming.client import AsyncClient
from wyoming.event import Event
from wyoming.ping import Ping, Pong
from wyoming.server import (
    AsyncEventHandler,
    AsyncServer,
    AsyncStdioServer,
    AsyncTcpServer,
    AsyncUnixServer,
)


class PingHandler(AsyncEventHandler):
    async def handle_event(self, event: Event) -> bool:
        if Ping.is_type(event.type):
            ping = Ping.from_event(event)
            await self.write_event(Pong(text=ping.text).event())
            return False

        return True


def test_from_uri() -> None:
    """Test AsyncServer.from_uri"""
    # Bad scheme
    with pytest.raises(ValueError):
        AsyncServer.from_uri("ftp://127.0.0.1:5000")

    # Missing hostname
    with pytest.raises(ValueError):
        AsyncServer.from_uri("tcp://:5000")

    # Missing port
    with pytest.raises(ValueError):
        AsyncServer.from_uri("tcp://127.0.0.1")

    stdio_server = AsyncServer.from_uri("stdio://")
    assert isinstance(stdio_server, AsyncStdioServer)

    tcp_server = AsyncServer.from_uri("tcp://127.0.0.1:5000")
    assert isinstance(tcp_server, AsyncTcpServer)
    assert tcp_server.host == "127.0.0.1"
    assert tcp_server.port == 5000

    unix_server = AsyncServer.from_uri("unix:///path/to/socket")
    assert isinstance(unix_server, AsyncUnixServer)
    assert unix_server.socket_path == Path("/path/to/socket")


@pytest.mark.asyncio
async def test_unix_server() -> None:
    """Test sending events to and from a Unix socket server."""
    with tempfile.TemporaryDirectory() as temp_dir:
        socket_path = Path(temp_dir) / "test.socket"
        uri = f"unix://{socket_path}"
        unix_server = AsyncServer.from_uri(uri)
        await unix_server.start(PingHandler)

        # Wait for path to exist
        while not socket_path.exists():
            await asyncio.sleep(0.1)

        client = AsyncClient.from_uri(uri)
        await client.connect()
        await client.write_event(Ping(text="test").event())

        event = await asyncio.wait_for(client.read_event(), timeout=1)
        assert event is not None
        assert Pong.is_type(event.type)
        assert Pong.from_event(event).text == "test"

        await client.disconnect()
        await unix_server.stop()


@pytest.mark.asyncio
async def test_tcp_server() -> None:
    """Test sending events to and from a TCP server."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    port = sock.getsockname()[1]
    sock.close()

    uri = f"tcp://127.0.0.1:{port}"

    tcp_server = AsyncServer.from_uri(uri)
    await tcp_server.start(PingHandler)

    # Wait for socket to open
    for _ in range(10):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(("127.0.0.1", port))
            break
        except ConnectionRefusedError:
            await asyncio.sleep(0.1)

    client = AsyncClient.from_uri(uri)
    await client.connect()
    await client.write_event(Ping(text="test").event())

    event = await asyncio.wait_for(client.read_event(), timeout=1)
    assert event is not None
    assert Pong.is_type(event.type)
    assert Pong.from_event(event).text == "test"

    await client.disconnect()
    await tcp_server.stop()


@pytest.mark.asyncio
async def test_tcp_server_sigterm() -> None:
    """Test that SIGTERM gracefully stops a running TCP server."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()

    tcp_server = AsyncServer.from_uri(f"tcp://127.0.0.1:{port}")
    run_task = asyncio.create_task(tcp_server.run(PingHandler))

    # Wait for socket to open
    for _ in range(10):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(("127.0.0.1", port))
            sock.close()
            break
        except ConnectionRefusedError:
            await asyncio.sleep(0.1)

    os.kill(os.getpid(), signal.SIGTERM)

    # run() must return without raising
    await asyncio.wait_for(run_task, timeout=5)


@pytest.mark.asyncio
async def test_tcp_server_cancel() -> None:
    """Test that external cancellation of run() still propagates."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()

    tcp_server = AsyncServer.from_uri(f"tcp://127.0.0.1:{port}")
    run_task = asyncio.create_task(tcp_server.run(PingHandler))
    await asyncio.sleep(0.1)

    run_task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await run_task



"""Test that eventable classes are correct."""

import importlib
import inspect
import pkgutil
from dataclasses import MISSING, fields
from typing import Any, Dict, List, Type

import pytest

import wyoming
from wyoming.audio import AudioFormat
from wyoming.event import Eventable
from wyoming.info import (
    AsrModel,
    AsrProgram,
    Attribution,
    HandleModel,
    HandleProgram,
    IntentModel,
    IntentProgram,
    MicProgram,
    Satellite,
    SndProgram,
    TtsProgram,
    TtsVoice,
    TtsVoiceSpeaker,
    WakeModel,
    WakeProgram,
)
from wyoming.intent import Entity
from wyoming.pipeline import PipelineStage
from wyoming.tts import SynthesizeVoice


def all_unique_eventables() -> List[Type[Eventable]]:
    found_classes = set()
    for _finder, name, _ispkg in pkgutil.walk_packages(
        wyoming.__path__, wyoming.__name__ + "."
    ):
        module = importlib.import_module(name)
        for _cls_name, cls_obj in inspect.getmembers(module, inspect.isclass):
            if issubclass(cls_obj, Eventable) and cls_obj is not Eventable:
                found_classes.add(cls_obj)  # set for uniqueness

    return list(found_classes)


EVENTABLE_CLASSES = all_unique_eventables()
NO_ARGS_CLASSES = [
    cls
    for cls in EVENTABLE_CLASSES
    if all(
        f.default is not MISSING or f.default_factory is not MISSING
        for f in fields(cls)  # type: ignore[arg-type]
    )
]

TEST_NAME = "test-name"
TEST_TEXT = "test text"
TEST_CONTEXT = {"test": "context"}
TEST_AUDIO_SETTINGS = {"rate": 22050, "width": 2, "channels": 1}
TEST_AUDIO_FORMAT = AudioFormat(22050, 2, 1)
TEST_ID = "test-id"
TEST_LANGUAGE = "test-language"
TEST_SPEAKER = "test-speaker"
TEST_TIMESTAMP = 1234
TEST_URL = "test://url"
TEST_DESCRIPTION = "test description"
TEST_VERSION = "test-version"
TEST_WAKE_WORD = "test-wake-word"
TEST_VOICE = SynthesizeVoice(name=TEST_NAME, speaker=TEST_SPEAKER)
TEST_ATTRIBUTION = Attribution(name=TEST_NAME, url=TEST_URL)

TEST_DATA: Dict[str, Dict[str, Any]] = {
    # info
    "Describe": {},
    "SelectProgram": {"name": TEST_NAME},
    "Info": {
        "asr": [
            AsrProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                models=[
                    AsrModel(
                        name=TEST_NAME,
                        attribution=TEST_ATTRIBUTION,
                        installed=True,
                        description=TEST_DESCRIPTION,
                        version=TEST_VERSION,
                        languages=[TEST_LANGUAGE],
                    )
                ],
                supports_transcript_streaming=True,
                requires_external_vad=False,
                prefers_auto_gain_enabled=False,
                prefers_noise_reduction_enabled=False,
            )
        ],
        "tts": [
            TtsProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                voices=[
                    TtsVoice(
                        name=TEST_NAME,
                        attribution=TEST_ATTRIBUTION,
                        installed=True,
                        description=TEST_DESCRIPTION,
                        version=TEST_VERSION,
                        languages=[TEST_LANGUAGE],
                        speakers=[TtsVoiceSpeaker(TEST_SPEAKER)],
                    )
                ],
                supports_synthesize_streaming=True,
            )
        ],
        "handle": [
            HandleProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                models=[
                    HandleModel(
                        name=TEST_NAME,
                        attribution=TEST_ATTRIBUTION,
                        installed=True,
                        description=TEST_DESCRIPTION,
                        version=TEST_VERSION,
                        languages=[TEST_LANGUAGE],
                    )
                ],
                supports_handled_streaming=True,
                supports_home_control=True,
            )
        ],
        "intent": [
            IntentProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                models=[
                    IntentModel(
                        name=TEST_NAME,
                        attribution=TEST_ATTRIBUTION,
                        installed=True,
                        description=TEST_DESCRIPTION,
                        version=TEST_VERSION,
                        languages=[TEST_LANGUAGE],
                    )
                ],
            )
        ],
        "wake": [
            WakeProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                models=[
                    WakeModel(
                        name=TEST_NAME,
                        attribution=TEST_ATTRIBUTION,
                        installed=True,
                        description=TEST_DESCRIPTION,
                        version=TEST_VERSION,
                        languages=[TEST_LANGUAGE],
                        phrase="test phrase",
                    )
                ],
            )
        ],
        "mic": [
            MicProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                mic_format=TEST_AUDIO_FORMAT,
            )
        ],
        "snd": [
            SndProgram(
                name=TEST_NAME,
                attribution=TEST_ATTRIBUTION,
                installed=True,
                description=TEST_DESCRIPTION,
                version=TEST_VERSION,
                snd_format=TEST_AUDIO_FORMAT,
            )
        ],
        "satellite": Satellite(
            name=TEST_NAME,
            attribution=TEST_ATTRIBUTION,
            installed=True,
            description=TEST_DESCRIPTION,
            version=TEST_VERSION,
            area="test area",
            has_vad=True,
            active_wake_words=[TEST_WAKE_WORD],
            max_active_wake_words=1,
            supports_trigger=True,
        ),
    },
    # audio
    "AudioStart": TEST_AUDIO_SETTINGS,
    "AudioChunk": {**TEST_AUDIO_SETTINGS, "audio": bytes(100)},
    "AudioStop": {},
    # vad
    "VoiceStarted": {"timestamp": TEST_TIMESTAMP},
    "VoiceStopped": {"timestamp": TEST_TIMESTAMP},
    # wake
    "Detect": {"names": [TEST_NAME], "context": TEST_CONTEXT},
    "Detection": {
        "name": TEST_NAME,
        "timestamp": TEST_TIMESTAMP,
        "speaker": TEST_SPEAKER,
        "context": TEST_CONTEXT,
    },
    "NotDetected": {"context": TEST_CONTEXT},
    # asr
    "Transcribe": {
        "name": TEST_NAME,
        "context": TEST_CONTEXT,
        "language": TEST_LANGUAGE,
        "vad_sensitivity": "default",
        "transcript_names": ["EcoBee"],
        "transcript_terms": ["pause"],
    },
    "Transcript": {
        "text": TEST_TEXT,
        "context": TEST_CONTEXT,
        "language": TEST_LANGUAGE,
    },
    "TranscriptStart": {
        "context": TEST_CONTEXT,
        "language": TEST_LANGUAGE,
    },
    "TranscriptChunk": {"text": TEST_TEXT},
    "TranscriptStop": {},
    # intent
    "Recognize": {"text": TEST_TEXT, "context": TEST_CONTEXT},
    "Intent": {
        "name": "TestIntent",
        "entities": [Entity("test entity", "test-value")],
        "context": TEST_CONTEXT,
    },
    "NotRecognized": {"text": TEST_TEXT, "context": TEST_CONTEXT},
    "IntentsStart": {"context": TEST_CONTEXT},
    "IntentsStop": {"context": TEST_CONTEXT},
    # handle
    "Handled": {
        "text": TEST_TEXT,
        "context": TEST_CONTEXT,
    },
    "HandledStart": {"context": TEST_CONTEXT},
    "HandledChunk": {"text": TEST_TEXT},
    "HandledStop": {},
    "NotHandled": {
        "text": TEST_TEXT,
        "context": TEST_CONTEXT,
    },
    # tts
    "SynthesizeStart": {
        "voice": TEST_VOICE,
        "text_format": "ssml",
        "context": TEST_CONTEXT,
    },
    "SynthesizeChunk": {"text": TEST_TEXT},
    "SynthesizeStop": {},
    "SynthesizeStopped": {},
    "Synthesize": {
        "text": TEST_TEXT,
        "voice": TEST_VOICE,
        "text_format": "text",
        "context": TEST_CONTEXT,
    },
    # timers
    "TimerStarted": {"id": TEST_ID, "total_seconds": 100},
    "TimerUpdated": {"id": TEST_ID, "total_seconds": 100, "is_active": True},
    "TimerFinished": {"id": TEST_ID},
    "TimerCancelled": {"id": TEST_ID},
    # snd
    "Played": {},
    # satellite
    "RunSatellite": {},
    "PauseSatellite": {},
    "StreamingStarted": {},
    "StreamingStopped": {},
    "SatelliteConnected": {},
    "SatelliteDisconnected": {},
    # misc
    "Error": {"text": TEST_TEXT},
    "Ping": {},
    "Pong": {},
    "RunPipeline": {"start_stage": PipelineStage.ASR, "end_stage": PipelineStage.TTS},
    "UserEvent": {
        "name": TEST_NAME,
        "data": {TEST_ID: TEST_TEXT},
        "context": TEST_CONTEXT,
    },
}


@pytest.mark.parametrize("cls", EVENTABLE_CLASSES)
def test_eventable_round_trip(cls: Type[Eventable]) -> None:
    init_kwargs = TEST_DATA[cls.__name__]
    instance = cls(**init_kwargs)

    # Test event() method
    event = instance.event()
    assert event.type is not None, f"{cls} returned event with no type"

    # Test is_type matches event.type
    assert cls.is_type(event.type), f"{cls}.is_type failed for {event.type}"

    # Test from_event returns an equivalent object
    round_trip = cls.from_event(event)
    assert round_trip == instance, f"{cls}.from_event failed to round-trip {instance}"


@pytest.mark.parametrize("cls", NO_ARGS_CLASSES)
def test_eventable_no_args(cls: Type[Eventable]) -> None:
    instance = cls()

    # Test event() method
    event = instance.event()
    assert event.type is not None, f"{cls} returned event with no type"

    # Test is_type matches event.type
    assert cls.is_type(event.type), f"{cls}.is_type failed for {event.type}"

    # Test from_event returns an equivalent object
    round_trip = cls.from_event(event)
    assert round_trip == instance, f"{cls}.from_event failed to round-trip {instance}"



"""HTTP server for intent recognition/handling."""

import logging
from pathlib import Path
from typing import Any, Dict, List, Union

from flask import request
from werkzeug.exceptions import BadRequest

from wyoming.asr import Transcript
from wyoming.error import Error
from wyoming.handle import Handled, NotHandled
from wyoming.intent import Intent, IntentsStart, IntentsStop, NotRecognized

from .shared import check_args, get_app, get_argument_parser, get_client

_DIR = Path(__file__).parent
CONF_PATH = _DIR / "conf" / "intent.yaml"


def main():
    parser = get_argument_parser()
    parser.add_argument("--language", help="Language for text")
    args = parser.parse_args()
    check_args(parser, args)
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    app = get_app("intent", CONF_PATH, args)

    @app.route("/api/recognize-intent", methods=["POST", "GET"])
    async def api_recognize_intent() -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        if request.method == "POST":
            text = request.data.decode()
        else:
            text = request.args.get("text", "")

        if not text:
            raise BadRequest("Text is required")

        language = request.args.get("language", args.language)

        async with get_client(args) as client:
            await client.write_event(Transcript(text=text, language=language).event())

            type_name = "unknown"
            results: List[Dict[str, Any]] = []
            is_intent_list = False

            while True:
                event = await client.read_event()
                if event is None:
                    raise RuntimeError("Client disconnected")

                if IntentsStart.is_type(event.type):
                    is_intent_list = True
                    continue

                if Intent.is_type(event.type):
                    type_name = "intent"
                    intent = Intent.from_event(event)
                    results.append(intent.to_dict())
                    if not is_intent_list:
                        break

                if IntentsStop.is_type(event.type):
                    break

                if Handled.is_type(event.type):
                    type_name = "handled"
                    handled = Handled.from_event(event)
                    results.append(handled.to_dict())
                    break

                if NotRecognized.is_type(event.type):
                    type_name = "not-recognized"
                    not_recognized = NotRecognized.from_event(event)
                    results.append(not_recognized.to_dict())
                    break

                if NotHandled.is_type(event.type):
                    type_name = "not-handled"
                    not_handled = NotHandled.from_event(event)
                    results.append(not_handled.to_dict())
                    break

                if Error.is_type(event.type):
                    error = Error.from_event(event)
                    raise RuntimeError(
                        f"Unexpected error from client: code={error.code}, text={error.text}"
                    )

            if len(results) == 0:
                return {"success": False, "type": "unknown", "result": {}}

            if len(results) == 1:
                return {"success": True, "type": type_name, "result": results[0]}

            return {"success": True, "type": type_name, "result": results}

    app.run(args.host, args.port)


if __name__ == "__main__":
    main()



"""Ping/pong messages."""

from dataclasses import dataclass
from typing import Optional

from .event import Event, Eventable

_PING_TYPE = "ping"
_PONG_TYPE = "pong"


@dataclass
class Ping(Eventable):
    """Request pong message."""

    text: Optional[str] = None
    """Text to copy to response."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _PING_TYPE

    def event(self) -> Event:
        return Event(
            type=_PING_TYPE,
            data={"text": self.text},
        )

    @staticmethod
    def from_event(event: Event) -> "Ping":
        return Ping(text=event.data.get("text"))


@dataclass
class Pong(Eventable):
    """Response to ping message."""

    text: Optional[str] = None
    """Text copied from request."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _PONG_TYPE

    def event(self) -> Event:
        return Event(
            type=_PONG_TYPE,
            data={"text": self.text},
        )

    @staticmethod
    def from_event(event: Event) -> "Pong":
        return Pong(text=event.data.get("text"))



"""Speech to text."""

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from .event import Event, Eventable

DOMAIN = "asr"
_TRANSCRIPT_TYPE = "transcript"
_TRANSCRIBE_TYPE = "transcribe"
_TRANSCRIPT_START_TYPE = "transcript-start"
_TRANSCRIPT_CHUNK_TYPE = "transcript-chunk"
_TRANSCRIPT_STOP_TYPE = "transcript-stop"


class VadSensitivity(str, Enum):
    """How quickly the end of a voice command is detected."""

    DEFAULT = "default"
    RELAXED = "relaxed"
    AGGRESSIVE = "aggressive"


@dataclass
class Transcript(Eventable):
    """Transcription response from ASR system"""

    text: str
    """Text transcription of spoken audio"""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    language: Optional[str] = None
    """Language of the text."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _TRANSCRIPT_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {"text": self.text}
        if self.language is not None:
            data["language"] = self.language

        if self.context is not None:
            data["context"] = self.context

        return Event(type=_TRANSCRIPT_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "Transcript":
        return Transcript(
            text=event.data["text"],
            language=event.data.get("language"),
            context=event.data.get("context"),
        )


@dataclass
class Transcribe(Eventable):
    """Transcription request to ASR system.

    Followed by AudioStart, AudioChunk+, AudioStop
    """

    name: Optional[str] = None
    """Name of ASR model to use"""

    language: Optional[str] = None
    """Language of spoken audio to follow"""

    context: Optional[Dict[str, Any]] = None
    """Context from previous interactions."""

    vad_sensitivity: Optional[Union[str, VadSensitivity]] = None
    """How quickly the end of a voice command is detected."""

    transcript_names: Optional[List[str]] = None
    """Unique names that should bias the ASR transcription.

    These take priority over terms.
    """

    transcript_terms: Optional[List[str]] = None
    """Domain-specific terms that should bias the ASR transcription.

    These are lower priority than names.
    """

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _TRANSCRIBE_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {}
        if self.name is not None:
            data["name"] = self.name

        if self.language is not None:
            data["language"] = self.language

        if self.context is not None:
            data["context"] = self.context

        if isinstance(self.vad_sensitivity, VadSensitivity):
            data["vad_sensitivity"] = self.vad_sensitivity.value
        elif self.vad_sensitivity is not None:
            data["vad_sensitivity"] = self.vad_sensitivity

        if self.transcript_names is not None:
            data["transcript_names"] = self.transcript_names

        if self.transcript_terms is not None:
            data["transcript_terms"] = self.transcript_terms

        return Event(type=_TRANSCRIBE_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "Transcribe":
        data = event.data or {}

        vad_sensitivity: Optional[Union[str, VadSensitivity]] = None
        vad_sensitivity_str = data.get("vad_sensitivity")
        if vad_sensitivity_str is not None:
            try:
                vad_sensitivity = VadSensitivity(vad_sensitivity_str)
            except ValueError:
                vad_sensitivity = vad_sensitivity_str

        return Transcribe(
            name=data.get("name"),
            language=data.get("language"),
            context=data.get("context"),
            vad_sensitivity=vad_sensitivity,
            transcript_names=data.get("transcript_names"),
            transcript_terms=data.get("transcript_terms"),
        )


@dataclass
class TranscriptStart(Eventable):
    """Streaming transcription response from ASR system"""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    language: Optional[str] = None
    """Language of the text."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _TRANSCRIPT_START_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {}
        if self.language is not None:
            data["language"] = self.language
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_TRANSCRIPT_START_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "TranscriptStart":
        if not event.data:
            return TranscriptStart()

        return TranscriptStart(
            context=event.data.get("context"), language=event.data.get("language")
        )


@dataclass
class TranscriptChunk(Eventable):
    """Chunk of streaming transcription response from ASR system"""

    text: str
    """Chunk of transcript text."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _TRANSCRIPT_CHUNK_TYPE

    def event(self) -> Event:
        return Event(type=_TRANSCRIPT_CHUNK_TYPE, data={"text": self.text})

    @staticmethod
    def from_event(event: Event) -> "TranscriptChunk":
        return TranscriptChunk(text=event.data["text"])


@dataclass
class TranscriptStop(Eventable):
    """End of streaming transcription response from ASR system"""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _TRANSCRIPT_STOP_TYPE

    def event(self) -> Event:
        return Event(type=_TRANSCRIPT_STOP_TYPE)

    @staticmethod
    def from_event(event: Event) -> "TranscriptStop":
        return TranscriptStop()



"""HTTP server for wake word detection."""

import io
import logging
import wave
from pathlib import Path
from typing import Set

from flask import Response, jsonify, request

from wyoming.audio import wav_to_chunks
from wyoming.error import Error
from wyoming.wake import Detect, Detection, NotDetected

from .shared import check_args, get_app, get_argument_parser, get_client

_DIR = Path(__file__).parent
CONF_PATH = _DIR / "conf" / "wake.yaml"


def main():
    parser = get_argument_parser()
    parser.add_argument("--wake-word-name", action="append")
    parser.add_argument("--samples-per-chunk", type=int, default=1024)
    args = parser.parse_args()
    check_args(parser, args)
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    app = get_app("wake", CONF_PATH, args)

    @app.route("/api/detect-wake-word", methods=["POST", "GET"])
    async def api_wake() -> Response:
        wake_word_names: Set[str] = set()

        async with get_client(args) as client:
            if args.wake_word_name:
                # From command-line
                wake_word_names.update(args.wake_word_name)

            wake_word_names.update(request.args.getlist("wake_words"))

            if wake_word_names:
                await client.write_event(Detect(list(wake_word_names)).event())

            with io.BytesIO(request.data) as wav_io:
                with wave.open(wav_io, "rb") as wav_file:
                    chunks = wav_to_chunks(
                        wav_file,
                        samples_per_chunk=args.samples_per_chunk,
                        start_event=True,
                        stop_event=True,
                    )
                    for chunk in chunks:
                        await client.write_event(chunk.event())

            while True:
                event = await client.read_event()
                if event is None:
                    raise RuntimeError("Client disconnected")

                if Detection.is_type(event.type) or NotDetected.is_type(event.type):
                    return jsonify(event.to_dict())

                if Error.is_type(event.type):
                    error = Error.from_event(event)
                    raise RuntimeError(
                        f"Unexpected error from client: code={error.code}, text={error.text}"
                    )

    app.run(args.host, args.port)


if __name__ == "__main__":
    main()



"""Pipeline events."""

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

from .event import Event, Eventable

_RUN_PIPELINE_TYPE = "run-pipeline"


class PipelineStage(str, Enum):
    """Stages of a pipeline."""

    WAKE = "wake"
    """Wake word detection."""

    ASR = "asr"
    """Speech-to-text (a.k.a. automated speech recognition)."""

    INTENT = "intent"
    """Intent recognition."""

    HANDLE = "handle"
    """Intent handling."""

    TTS = "tts"
    """Text-to-speech."""


@dataclass
class RunPipeline(Eventable):
    """Run a pipeline"""

    start_stage: PipelineStage
    """Stage to start the pipeline on."""

    end_stage: PipelineStage
    """Stage to end the pipeline on."""

    wake_word_name: Optional[str] = None
    """Name of wake word that triggered this pipeline."""

    restart_on_end: bool = False
    """True if pipeline should restart automatically after ending."""

    wake_word_names: Optional[List[str]] = None
    """Wake word names to listen for (start_stage = wake)."""

    announce_text: Optional[str] = None
    """Text to announce using text-to-speech (start_stage = tts)"""

    def __post_init__(self) -> None:
        start_valid = True
        end_valid = True

        if self.start_stage == PipelineStage.WAKE:
            if self.end_stage not in (
                PipelineStage.WAKE,
                PipelineStage.ASR,
                PipelineStage.INTENT,
                PipelineStage.HANDLE,
                PipelineStage.TTS,
            ):
                end_valid = False
        elif self.start_stage == PipelineStage.ASR:
            if self.end_stage not in (
                PipelineStage.ASR,
                PipelineStage.INTENT,
                PipelineStage.HANDLE,
                PipelineStage.TTS,
            ):
                end_valid = False
        elif self.start_stage == PipelineStage.INTENT:
            if self.end_stage not in (
                PipelineStage.INTENT,
                PipelineStage.HANDLE,
                PipelineStage.TTS,
            ):
                end_valid = False
        elif self.start_stage == PipelineStage.HANDLE:
            if self.end_stage not in (
                PipelineStage.HANDLE,
                PipelineStage.TTS,
            ):
                end_valid = False
        elif self.start_stage == PipelineStage.TTS:
            if self.end_stage not in (PipelineStage.TTS,):
                end_valid = False
        else:
            start_valid = False

        if not start_valid:
            raise ValueError(f"Invalid start stage: {self.start_stage}")

        if not end_valid:
            raise ValueError(f"Invalid end stage: {self.end_stage}")

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _RUN_PIPELINE_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {
            "start_stage": self.start_stage.value,
            "end_stage": self.end_stage.value,
            "restart_on_end": self.restart_on_end,
        }

        if self.wake_word_name is not None:
            data["wake_word_name"] = self.wake_word_name

        if self.wake_word_names:
            data["wake_word_names"] = self.wake_word_names

        if self.announce_text is not None:
            data["announce_text"] = self.announce_text

        return Event(type=_RUN_PIPELINE_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "RunPipeline":
        return RunPipeline(
            start_stage=PipelineStage(event.data["start_stage"]),
            end_stage=PipelineStage(event.data["end_stage"]),
            wake_word_name=event.data.get("wake_word_name"),
            restart_on_end=event.data.get("restart_on_end", False),
            wake_word_names=event.data.get("wake_word_names"),
            announce_text=event.data.get("announce_text"),
        )



"""Test audio utilities."""

import io
import wave

from wyoming.audio import AudioChunk, AudioChunkConverter, wav_to_chunks


def test_chunk_converter() -> None:
    """Test audio chunk converter."""
    converter = AudioChunkConverter(rate=16000, width=2, channels=1)
    input_chunk = AudioChunk(
        rate=48000,
        width=4,
        channels=2,
        audio=bytes(1 * 48000 * 4 * 2),  # 1 sec
    )

    output_chunk = converter.convert(input_chunk)
    assert output_chunk.rate == 16000
    assert output_chunk.width == 2
    assert output_chunk.channels == 1
    assert len(output_chunk.audio) == 1 * 16000 * 2 * 1  # 1 sec


def test_chunk_converter_8bit() -> None:
    """Test conversion to/from unsigned 8-bit audio.

    WAV 8-bit samples are unsigned (silence = 128), unlike wider samples which
    are signed (silence = 0). The converter must account for this.
    """
    # Unsigned 8-bit: silence, max, min, and a positive value
    audio_8bit = bytes([128, 255, 0, 192])

    # 8-bit -> 16-bit: silence must map to 0, not full-scale noise
    to_16bit = AudioChunkConverter(width=2)
    chunk_16bit = to_16bit.convert(
        AudioChunk(rate=16000, width=1, channels=1, audio=audio_8bit)
    )
    assert chunk_16bit.width == 2
    samples_16bit = [
        int.from_bytes(chunk_16bit.audio[i : i + 2], "little", signed=True)
        for i in range(0, len(chunk_16bit.audio), 2)
    ]
    assert samples_16bit[0] == 0  # unsigned silence -> signed silence

    # 16-bit -> 8-bit round-trips back to the original unsigned bytes
    to_8bit = AudioChunkConverter(width=1)
    chunk_8bit = to_8bit.convert(chunk_16bit)
    assert chunk_8bit.width == 1
    assert chunk_8bit.audio == audio_8bit


def test_wav_to_chunks() -> None:
    """Test WAV file to audio chunks."""
    with io.BytesIO() as wav_io:
        wav_write: wave.Wave_write = wave.open(wav_io, "wb")
        with wav_write:
            wav_write.setframerate(16000)
            wav_write.setsampwidth(2)
            wav_write.setnchannels(1)
            wav_write.writeframes(bytes(1 * 16000 * 2 * 1))  # 1 sec

        wav_io.seek(0)
        wav_bytes = wav_io.getvalue()

    with io.BytesIO(wav_bytes) as wav_io:
        wav_read: wave.Wave_read = wave.open(wav_io, "rb")
        chunks = list(wav_to_chunks(wav_read, samples_per_chunk=1000))
        assert len(chunks) == 16
        for chunk in chunks:
            assert isinstance(chunk, AudioChunk)
            assert chunk.rate == 16000
            assert chunk.width == 2
            assert chunk.channels == 1
            assert len(chunk.audio) == 1000 * 2  # 1000 samples



"""Satellite events."""

from dataclasses import dataclass

from .event import Event, Eventable

_RUN_SATELLITE_TYPE = "run-satellite"
_PAUSE_SATELLITE_TYPE = "pause-satellite"
_STREAMING_STARTED_TYPE = "streaming-started"
_STREAMING_STOPPED_TYPE = "streaming-stopped"
_SATELLITE_CONNECTED_TYPE = "satellite-connected"
_SATELLITE_DISCONNECTED_TYPE = "satellite-disconnected"


@dataclass
class RunSatellite(Eventable):
    """Informs the satellite that the server is ready to run a pipeline."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _RUN_SATELLITE_TYPE

    def event(self) -> Event:
        return Event(type=_RUN_SATELLITE_TYPE)

    @staticmethod
    def from_event(event: Event) -> "RunSatellite":
        return RunSatellite()


@dataclass
class PauseSatellite(Eventable):
    """Informs the satellite that the server is not ready to run a pipeline."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _PAUSE_SATELLITE_TYPE

    def event(self) -> Event:
        return Event(type=_PAUSE_SATELLITE_TYPE)

    @staticmethod
    def from_event(event: Event) -> "PauseSatellite":
        return PauseSatellite()


@dataclass
class StreamingStarted(Eventable):
    """Satellite has started streaming audio to server."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _STREAMING_STARTED_TYPE

    def event(self) -> Event:
        return Event(type=_STREAMING_STARTED_TYPE)

    @staticmethod
    def from_event(event: Event) -> "StreamingStarted":
        return StreamingStarted()


@dataclass
class StreamingStopped(Eventable):
    """Satellite has stopped streaming audio to server."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _STREAMING_STOPPED_TYPE

    def event(self) -> Event:
        return Event(type=_STREAMING_STOPPED_TYPE)

    @staticmethod
    def from_event(event: Event) -> "StreamingStopped":
        return StreamingStopped()


@dataclass
class SatelliteConnected(Eventable):
    """Satellite has connected to server."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _SATELLITE_CONNECTED_TYPE

    def event(self) -> Event:
        return Event(type=_SATELLITE_CONNECTED_TYPE)

    @staticmethod
    def from_event(event: Event) -> "SatelliteConnected":
        return SatelliteConnected()


@dataclass
class SatelliteDisconnected(Eventable):
    """Satellite has disconnected from server."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _SATELLITE_DISCONNECTED_TYPE

    def event(self) -> Event:
        return Event(type=_SATELLITE_DISCONNECTED_TYPE)

    @staticmethod
    def from_event(event: Event) -> "SatelliteDisconnected":
        return SatelliteDisconnected()



"""Intent recognition and handling."""

from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Optional

from .event import Event, Eventable

DOMAIN = "intent"
_RECOGNIZE_TYPE = "recognize"
_INTENT_TYPE = "intent"
_NOT_RECOGNIZED_TYPE = "not-recognized"
_INTENTS_START_TYPE = "intents-start"
_INTENTS_STOP_TYPE = "intents-stop"


@dataclass
class Entity:
    """Named entity with a value."""

    name: str
    value: Optional[Any] = None


@dataclass
class Recognize(Eventable):
    """Request to recognize an event from text."""

    text: str
    """Text with intent in natural language."""

    context: Optional[Dict[str, Any]] = None
    """Context from previous interactions."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _RECOGNIZE_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {"text": self.text}
        if self.context is not None:
            data["context"] = self.context
        return Event(type=_RECOGNIZE_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "Recognize":
        return Recognize(text=event.data["text"], context=event.data.get("context"))


@dataclass
class Intent(Eventable):
    """Result of successful intent recognition."""

    name: str
    """Name of intent."""

    entities: List[Entity] = field(default_factory=list)
    """Named entities with values."""

    text: Optional[str] = None
    """Human-readable response."""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _INTENT_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {"name": self.name}
        if self.entities:
            data["entities"] = [asdict(entity) for entity in self.entities]
        if self.text is not None:
            data["text"] = self.text
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_INTENT_TYPE, data=data)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "Intent":
        entity_dicts = data.get("entities")
        if entity_dicts:
            entities: List[Entity] = [
                Entity(**entity_dict) for entity_dict in entity_dicts
            ]
        else:
            entities = []

        return Intent(
            name=data["name"],
            entities=entities,
            text=data.get("text"),
            context=data.get("context"),
        )

    @staticmethod
    def from_event(event: Event) -> "Intent":
        return Intent.from_dict(event.data)

    def to_rhasspy(self) -> Dict[str, Any]:
        return {
            "intent": {
                "name": self.name,
            },
            "entities": [
                {"entity": entity.name, "value": entity.value}
                for entity in self.entities
            ],
            "slots": {entity.name: entity.value for entity in self.entities},
        }


@dataclass
class NotRecognized(Eventable):
    """Result of intent recognition failure."""

    text: Optional[str] = None
    """Human-readable response."""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _NOT_RECOGNIZED_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {}
        if self.text is not None:
            data["text"] = self.text
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_NOT_RECOGNIZED_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "NotRecognized":
        return NotRecognized(
            text=event.data.get("text"), context=event.data.get("context")
        )


@dataclass
class IntentsStart(Eventable):
    """Start of one or more intents.

    Event flow:
    intents-start
    intent
    [intent]...
    intents-stop
    """

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _INTENTS_START_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {}
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_INTENTS_START_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "IntentsStart":
        if not event.data:
            return IntentsStart()

        return IntentsStart(context=event.data.get("context"))


@dataclass
class IntentsStop(Eventable):
    """End of intent series.

    Event flow:
    intents-start
    intent
    [intent]...
    intents-stop
    """

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _INTENTS_STOP_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {}
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_INTENTS_STOP_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "IntentsStop":
        if not event.data:
            return IntentsStop()

        return IntentsStop(context=event.data.get("context"))



"""Voice activity detection."""

from dataclasses import dataclass
from typing import Optional

from .event import Event, Eventable

DOMAIN = "vad"
_STARTED_TYPE = "voice-started"
_STOPPED_TYPE = "voice-stopped"


@dataclass
class VoiceStarted(Eventable):
    """User has started speaking."""

    timestamp: Optional[int] = None
    """Milliseconds"""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _STARTED_TYPE

    def event(self) -> Event:
        return Event(
            type=_STARTED_TYPE,
            data={"timestamp": self.timestamp},
        )

    @staticmethod
    def from_event(event: Event) -> "VoiceStarted":
        return VoiceStarted(timestamp=event.data.get("timestamp"))


@dataclass
class VoiceStopped(Eventable):
    """User has stopped speaking."""

    timestamp: Optional[int] = None
    """Milliseconds"""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _STOPPED_TYPE

    def event(self) -> Event:
        return Event(
            type=_STOPPED_TYPE,
            data={"timestamp": self.timestamp},
        )

    @staticmethod
    def from_event(event: Event) -> "VoiceStopped":
        return VoiceStopped(timestamp=event.data.get("timestamp"))



"""Shared code for HTTP servers.

These servers are unauthenticated by design and are intended to run on a
trusted network only. See SECURITY.md.
"""

import argparse
import logging
from pathlib import Path
from typing import Union
from urllib.parse import urlparse

from flask import Flask, jsonify, redirect, request
from swagger_ui import flask_api_doc  # pylint: disable=no-name-in-module
from werkzeug.exceptions import BadRequest, HTTPException

from wyoming.client import AsyncClient
from wyoming.info import Describe, Info

_LOGGER = logging.getLogger(__name__)

# Schemes a request is allowed to ask for. 'stdio://' is excluded because it is
# meaningless for an HTTP server and would consume the server's own stdin.
_REQUEST_URI_SCHEMES = ("tcp", "unix")


def get_argument_parser() -> argparse.ArgumentParser:
    """Create argument parser with shared arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", default="0.0.0.0")
    parser.add_argument("--port", type=int, default=5000)
    parser.add_argument(
        "--uri", help="URI of Wyoming service (required without --allow-uri-override)"
    )
    parser.add_argument(
        "--allow-uri-override",
        action="store_true",
        help="Allow each request to select the Wyoming service with the 'uri' "
        "query parameter. Any client can then make this server connect to any "
        "address it can reach, so only use this on a trusted network.",
    )
    parser.add_argument(
        "--connect-timeout",
        type=float,
        default=10.0,
        help="Seconds to wait when connecting to a Wyoming service (default: 10)",
    )
    parser.add_argument(
        "--read-timeout",
        type=float,
        default=300.0,
        help="Seconds to wait for each event from a Wyoming service. "
        "Applies per event, not to the request as a whole. "
        "Use 0 to wait indefinitely (default: 300)",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG logs to console"
    )
    return parser


def check_args(parser: argparse.ArgumentParser, args: argparse.Namespace) -> None:
    """Verify that a Wyoming service can be determined for incoming requests."""
    if not (args.uri or args.allow_uri_override):
        parser.error(
            "--uri is required, or use --allow-uri-override to let each request "
            "select the Wyoming service (only do this on a trusted network)"
        )


def get_uri(args: argparse.Namespace) -> str:
    """Determine which Wyoming service the current request should use.

    A request may only choose the service if the operator opted in with
    --allow-uri-override. Otherwise the configured --uri is always used.
    """
    request_uri = request.args.get("uri")
    if request_uri:
        if not args.allow_uri_override:
            raise BadRequest(
                "This server does not allow requests to select the Wyoming "
                "service. It must be started with --allow-uri-override."
            )

        if urlparse(request_uri).scheme not in _REQUEST_URI_SCHEMES:
            raise BadRequest(
                "uri must be one of: "
                + ", ".join(f"{scheme}://" for scheme in _REQUEST_URI_SCHEMES)
            )

        return request_uri

    if not args.uri:
        raise BadRequest("uri is required")

    return args.uri


def get_client(args: argparse.Namespace) -> AsyncClient:
    """Create a Wyoming client for the current request."""
    return AsyncClient.from_uri(
        get_uri(args),
        # 0 disables the timeout. Without one, a service that accepts the
        # connection but never responds occupies the worker indefinitely.
        connect_timeout=args.connect_timeout or None,
        read_timeout=args.read_timeout or None,
    )


def get_app(
    name: str, openapi_config_path: Union[str, Path], args: argparse.Namespace
) -> Flask:
    """Create Flask app with default endpoints."""
    if not (args.uri or args.allow_uri_override):
        raise ValueError(
            "Either 'uri' must be set, or 'allow_uri_override' must be enabled"
        )

    app = Flask(name)

    @app.route("/")
    def redirect_to_api():
        return redirect("/api")

    @app.route("/api/info", methods=["GET"])
    async def api_info():
        async with get_client(args) as client:
            await client.write_event(Describe().event())

            while True:
                event = await client.read_event()
                if event is None:
                    raise RuntimeError("Client disconnected")

                if Info.is_type(event.type):
                    info = Info.from_event(event)
                    return jsonify(info.to_dict())

    @app.errorhandler(Exception)
    async def handle_error(err):
        """Return error as text."""
        if isinstance(err, HTTPException):
            # Raised deliberately, so the description is safe to return
            return (f"{err.name}: {err.description}", err.code)

        # Log server-side only: the message may describe hosts and services that
        # the client should not learn about.
        _LOGGER.exception("Unexpected error while handling request")
        return ("Internal Server Error", 500)

    flask_api_doc(
        app, config_path=str(openapi_config_path), url_prefix="/api", title="API doc"
    )

    return app



"""Text to speech."""

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Union

from .event import Event, Eventable

DOMAIN = "tts"
_SYNTHESIZE_TYPE = "synthesize"

# streaming
_SYNTHESIZE_START_TYPE = "synthesize-start"
_SYNTHESIZE_CHUNK_TYPE = "synthesize-chunk"
_SYNTHESIZE_STOP_TYPE = "synthesize-stop"
_SYNTHESIZE_STOPPED_TYPE = "synthesize-stopped"


class SynthesizeTextFormat(str, Enum):
    """Format of text to synthesize."""

    TEXT = "text"
    SSML = "ssml"


@dataclass
class SynthesizeVoice:
    """Information about the desired voice for synthesis."""

    name: Optional[str] = None
    """Voice name from tts info (overrides language)."""

    language: Optional[str] = None
    """Voice language from tts info."""

    speaker: Optional[str] = None
    """Voice speaker from tts info."""

    def to_dict(self) -> Dict[str, str]:
        if self.name is not None:
            voice = {"name": self.name}
            if self.speaker is not None:
                voice["speaker"] = self.speaker
        elif self.language is not None:
            voice = {"language": self.language}
        else:
            voice = {}

        return voice

    @staticmethod
    def from_dict(voice: Dict[str, Any]) -> "Optional[SynthesizeVoice]":
        if "name" in voice:
            return SynthesizeVoice(
                name=voice["name"],
                speaker=voice.get("speaker"),
            )

        if "language" in voice:
            return SynthesizeVoice(language=voice["language"])

        return None


@dataclass
class Synthesize(Eventable):
    """Request to synthesize audio from text."""

    text: str
    """Text to synthesize."""

    voice: Optional[SynthesizeVoice] = None
    """Voice to use during synthesis."""

    text_format: Optional[Union[str, SynthesizeTextFormat]] = None
    """Format of text."""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _SYNTHESIZE_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {"text": self.text}
        if self.voice is not None:
            data["voice"] = self.voice.to_dict()
        if isinstance(self.text_format, SynthesizeTextFormat):
            data["text_format"] = self.text_format.value
        elif self.text_format is not None:
            data["text_format"] = self.text_format
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_SYNTHESIZE_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "Synthesize":
        text_format: Optional[Union[str, SynthesizeTextFormat]] = event.data.get(
            "text_format"
        )
        if text_format is not None:
            try:
                text_format = SynthesizeTextFormat(text_format)
            except ValueError:
                # Leave original value
                pass

        return Synthesize(
            text=event.data["text"],
            voice=SynthesizeVoice.from_dict(event.data.get("voice", {})),
            text_format=text_format,
            context=event.data.get("context"),
        )


@dataclass
class SynthesizeStart(Eventable):
    """Start of streaming request to synthesize audio from text."""

    voice: Optional[SynthesizeVoice] = None
    """Voice to use during synthesis."""

    text_format: Optional[Union[str, SynthesizeTextFormat]] = None
    """Format of text."""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _SYNTHESIZE_START_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {}
        if self.voice is not None:
            data["voice"] = self.voice.to_dict()
        if isinstance(self.text_format, SynthesizeTextFormat):
            data["text_format"] = self.text_format.value
        elif self.text_format is not None:
            data["text_format"] = self.text_format
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_SYNTHESIZE_START_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "SynthesizeStart":
        text_format: Optional[Union[str, SynthesizeTextFormat]] = event.data.get(
            "text_format"
        )
        if text_format is not None:
            try:
                text_format = SynthesizeTextFormat(text_format)
            except ValueError:
                # Leave original value
                pass

        return SynthesizeStart(
            voice=SynthesizeVoice.from_dict(event.data.get("voice", {})),
            text_format=text_format,
            context=event.data.get("context"),
        )


@dataclass
class SynthesizeChunk(Eventable):
    """Text chunk from streaming request to synthesize audio from text."""

    text: str
    """Chunk of text to synthesize."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _SYNTHESIZE_CHUNK_TYPE

    def event(self) -> Event:
        return Event(type=_SYNTHESIZE_CHUNK_TYPE, data={"text": self.text})

    @staticmethod
    def from_event(event: Event) -> "SynthesizeChunk":
        return SynthesizeChunk(text=event.data["text"])


@dataclass
class SynthesizeStop(Eventable):
    """End of streaming request to synthesize audio from text."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _SYNTHESIZE_STOP_TYPE

    def event(self) -> Event:
        return Event(type=_SYNTHESIZE_STOP_TYPE)

    @staticmethod
    def from_event(event: Event) -> "SynthesizeStop":
        return SynthesizeStop()


@dataclass
class SynthesizeStopped(Eventable):
    """End of streaming response to streaming request."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _SYNTHESIZE_STOPPED_TYPE

    def event(self) -> Event:
        return Event(type=_SYNTHESIZE_STOPPED_TYPE)

    @staticmethod
    def from_event(event: Event) -> "SynthesizeStopped":
        return SynthesizeStopped()



import asyncio
import json
import os
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, BinaryIO, Dict, Iterable, Optional

from .version import __version__

_TYPE = "type"
_DATA = "data"
_DATA_LENGTH = "data_length"
_PAYLOAD_LENGTH = "payload_length"
_NEWLINE = "\n".encode()
_VERSION = "version"
_VERSION_NUMBER = __version__


@dataclass
class Event:
    type: str
    data: Dict[str, Any] = field(default_factory=dict)
    payload: Optional[bytes] = None

    def to_dict(self) -> Dict[str, Any]:
        return {_TYPE: self.type, _DATA: self.data}

    @staticmethod
    def from_dict(event_dict: Dict[str, Any]) -> "Event":
        return Event(type=event_dict["type"], data=event_dict.get("data", {}))


class Eventable(ABC):
    @abstractmethod
    def event(self) -> Event:
        pass

    @staticmethod
    @abstractmethod
    def is_type(event_type: str) -> bool:
        pass

    def to_dict(self) -> Dict[str, Any]:
        return self.event().data

    @staticmethod
    @abstractmethod
    def from_event(event: Event) -> "Eventable":
        pass


async def async_get_stdin(
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> asyncio.StreamReader:
    """Get StreamReader for stdin."""
    if loop is None:
        loop = asyncio.get_running_loop()

    reader = asyncio.StreamReader()
    await loop.connect_read_pipe(
        lambda: asyncio.StreamReaderProtocol(reader), sys.stdin
    )

    return reader


async def async_get_stdout(
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> asyncio.StreamWriter:
    """Get StreamWriter for stdout."""
    if loop is None:
        loop = asyncio.get_running_loop()

    writer_transport, writer_protocol = await loop.connect_write_pipe(
        lambda: asyncio.streams.FlowControlMixin(loop=loop),
        os.fdopen(sys.stdout.fileno(), "wb"),
    )
    return asyncio.streams.StreamWriter(writer_transport, writer_protocol, None, loop)


async def async_read_event(reader: asyncio.StreamReader) -> Optional[Event]:
    try:
        json_line = await reader.readline()
        if not json_line:
            return None

        event_dict = json.loads(json_line)
        data_length = event_dict.get(_DATA_LENGTH)
        if (data_length is not None) and (data_length > 0):
            # Merge data
            data_bytes = await reader.readexactly(data_length)
            data_dict = event_dict.get(_DATA, {})
            data_dict.update(json.loads(data_bytes))
            event_dict[_DATA] = data_dict

        payload_length = event_dict.get(_PAYLOAD_LENGTH)

        payload: Optional[bytes] = None
        if (payload_length is not None) and (payload_length > 0):
            payload = await reader.readexactly(payload_length)

        return Event(
            type=event_dict[_TYPE], data=event_dict.get(_DATA, {}), payload=payload
        )
    except (KeyboardInterrupt, ValueError):
        pass

    return None


async def async_write_event(event: Event, writer: asyncio.StreamWriter):
    event_dict: Dict[str, Any] = event.to_dict()
    event_dict[_VERSION] = _VERSION_NUMBER

    data_dict = event_dict.pop(_DATA, None)
    data_bytes: Optional[bytes] = None
    if data_dict:
        data_bytes = json.dumps(data_dict, ensure_ascii=False).encode("utf-8")
        event_dict[_DATA_LENGTH] = len(data_bytes)

    if event.payload:
        event_dict[_PAYLOAD_LENGTH] = len(event.payload)

    json_line = json.dumps(event_dict, ensure_ascii=False)

    try:
        writer.writelines((json_line.encode(), _NEWLINE))

        if data_bytes:
            writer.write(data_bytes)

        if event.payload:
            writer.write(event.payload)

        await writer.drain()
    except KeyboardInterrupt:
        pass


async def async_write_events(events: Iterable[Event], writer: asyncio.StreamWriter):
    try:
        await asyncio.gather(*(async_write_event(event, writer) for event in events))
    except KeyboardInterrupt:
        pass


def read_event(reader: Optional[BinaryIO] = None) -> Optional[Event]:
    if reader is None:
        reader = sys.stdin.buffer

    try:
        json_line = reader.readline()

        if not json_line:
            return None

        event_dict = json.loads(json_line)
        data_length = event_dict.get(_DATA_LENGTH)
        if (data_length is not None) and (data_length > 0):
            # Merge data
            data_bytes = reader.read(data_length)
            while len(data_bytes) < data_length:
                chunk = reader.read(data_length - len(data_bytes))
                if not chunk:
                    # Stream closed before all data was read
                    return None
                data_bytes += chunk

            data_dict = event_dict.get(_DATA, {})
            data_dict.update(json.loads(data_bytes))
            event_dict[_DATA] = data_dict

        payload_length = event_dict.get(_PAYLOAD_LENGTH)

        payload: Optional[bytes] = None
        if (payload_length is not None) and (payload_length > 0):
            payload = reader.read(payload_length)
            while len(payload) < payload_length:
                chunk = reader.read(payload_length - len(payload))
                if not chunk:
                    # Stream closed before all data was read
                    return None
                payload += chunk

        return Event(
            type=event_dict[_TYPE], data=event_dict.get(_DATA, {}), payload=payload
        )
    except (KeyboardInterrupt, ValueError):
        pass

    return None


def write_event(event: Event, writer: Optional[BinaryIO] = None):
    if writer is None:
        writer = sys.stdout.buffer

    event_dict: Dict[str, Any] = event.to_dict()
    event_dict[_VERSION] = _VERSION_NUMBER

    data_dict = event_dict.pop(_DATA, None)
    data_bytes: Optional[bytes] = None
    if data_dict:
        data_bytes = json.dumps(data_dict, ensure_ascii=False).encode("utf-8")
        event_dict[_DATA_LENGTH] = len(data_bytes)

    if event.payload:
        event_dict[_PAYLOAD_LENGTH] = len(event.payload)

    json_line = json.dumps(event_dict, ensure_ascii=False)

    try:
        writer.writelines((json_line.encode(), _NEWLINE))

        if data_bytes:
            writer.write(data_bytes)

        if event.payload:
            writer.write(event.payload)

        writer.flush()
    except KeyboardInterrupt:
        pass



from wyoming.event import Event
from wyoming.tts import Synthesize, SynthesizeStart, SynthesizeTextFormat


def test_text_format() -> None:
    assert Synthesize.from_event(
        Event("synthesize", {"text": "test text", "text_format": "ssml"})
    ) == Synthesize("test text", text_format=SynthesizeTextFormat.SSML)

    assert Synthesize.from_event(
        Event("synthesize", {"text": "test text", "text_format": "other-format"})
    ) == Synthesize("test text", text_format="other-format")

    assert SynthesizeStart.from_event(
        Event("synthesize", {"text_format": "ssml"})
    ) == SynthesizeStart(text_format=SynthesizeTextFormat.SSML)

    assert SynthesizeStart.from_event(
        Event("synthesize", {"text_format": "other-format"})
    ) == SynthesizeStart(text_format="other-format")



from wyoming import pyaudioop


def pack(width, data, endian="little"):
    return b"".join(v.to_bytes(width, endian, signed=True) for v in data)


def unpack(width, data, endian="little"):
    return [
        int.from_bytes(data[i : i + width], endian, signed=True)
        for i in range(0, len(data), width)
    ]


packs = {w: (lambda *data, width=w: pack(width, data)) for w in (1, 2, 3, 4)}
maxvalues = {w: (1 << (8 * w - 1)) - 1 for w in (1, 2, 3, 4)}
minvalues = {w: -1 << (8 * w - 1) for w in (1, 2, 3, 4)}

datas = {
    1: b"\x00\x12\x45\xbb\x7f\x80\xff",
    2: packs[2](0, 0x1234, 0x4567, -0x4567, 0x7FFF, -0x8000, -1),
    3: packs[3](0, 0x123456, 0x456789, -0x456789, 0x7FFFFF, -0x800000, -1),
    4: packs[4](0, 0x12345678, 0x456789AB, -0x456789AB, 0x7FFFFFFF, -0x80000000, -1),
}

INVALID_DATA = [
    (b"abc", 0),
    (b"abc", 2),
    (b"ab", 3),
    (b"abc", 4),
]


def test_bias() -> None:
    """Test adding a bias to samples."""
    for w in 1, 2, 4:
        for bias in 0, 1, -1, 127, -128, 0x7FFFFFFF, -0x80000000:
            assert pyaudioop.bias(b"", w, bias) == b""

    assert pyaudioop.bias(datas[1], 1, 1) == b"\x01\x13\x46\xbc\x80\x81\x00"
    assert pyaudioop.bias(datas[1], 1, -1) == b"\xff\x11\x44\xba\x7e\x7f\xfe"
    assert pyaudioop.bias(datas[1], 1, 0x7FFFFFFF) == b"\xff\x11\x44\xba\x7e\x7f\xfe"
    assert pyaudioop.bias(datas[1], 1, -0x80000000) == datas[1]

    assert pyaudioop.bias(datas[2], 2, 1) == packs[2](
        1, 0x1235, 0x4568, -0x4566, -0x8000, -0x7FFF, 0
    )
    assert pyaudioop.bias(datas[4], 4, 1) == packs[4](
        1, 0x12345679, 0x456789AC, -0x456789AA, -0x80000000, -0x7FFFFFFF, 0
    )

    # Accepts bytearray and memoryview like the other functions
    assert pyaudioop.bias(bytearray(datas[1]), 1, 1) == b"\x01\x13\x46\xbc\x80\x81\x00"
    assert pyaudioop.bias(memoryview(datas[1]), 1, 1) == b"\x01\x13\x46\xbc\x80\x81\x00"


def test_lin2lin() -> None:
    """Test sample width conversions."""
    for w in 1, 2, 4:
        assert pyaudioop.lin2lin(datas[w], w, w) == datas[w]
        assert pyaudioop.lin2lin(bytearray(datas[w]), w, w) == datas[w]
        assert pyaudioop.lin2lin(memoryview(datas[w]), w, w) == datas[w]

    assert pyaudioop.lin2lin(datas[1], 1, 2) == packs[2](
        0, 0x1200, 0x4500, -0x4500, 0x7F00, -0x8000, -0x100
    )
    assert pyaudioop.lin2lin(datas[1], 1, 4) == packs[4](
        0, 0x12000000, 0x45000000, -0x45000000, 0x7F000000, -0x80000000, -0x1000000
    )
    assert pyaudioop.lin2lin(datas[2], 2, 1) == b"\x00\x12\x45\xba\x7f\x80\xff"
    assert pyaudioop.lin2lin(datas[2], 2, 4) == packs[4](
        0, 0x12340000, 0x45670000, -0x45670000, 0x7FFF0000, -0x80000000, -0x10000
    )
    assert pyaudioop.lin2lin(datas[4], 4, 1) == b"\x00\x12\x45\xba\x7f\x80\xff"
    assert pyaudioop.lin2lin(datas[4], 4, 2) == packs[2](
        0, 0x1234, 0x4567, -0x4568, 0x7FFF, -0x8000, -1
    )


def test_lin2lin_big_endian() -> None:
    """Test sample width conversions with big-endian samples."""
    data2 = pack(2, (0, 0x1234, 0x4567, -0x4567, 0x7FFF, -0x8000, -1), endian="big")
    data4 = pack(
        4,
        (0, 0x12345678, 0x456789AB, -0x456789AB, 0x7FFFFFFF, -0x80000000, -1),
        endian="big",
    )

    assert pyaudioop.lin2lin(data2, 2, 1, endian="big") == (
        b"\x00\x12\x45\xba\x7f\x80\xff"
    )
    assert pyaudioop.lin2lin(data2, 2, 4, endian="big") == pack(
        4,
        (0, 0x12340000, 0x45670000, -0x45670000, 0x7FFF0000, -0x80000000, -0x10000),
        endian="big",
    )
    assert pyaudioop.lin2lin(data4, 4, 1, endian="big") == (
        b"\x00\x12\x45\xba\x7f\x80\xff"
    )
    assert pyaudioop.lin2lin(data4, 4, 2, endian="big") == pack(
        2, (0, 0x1234, 0x4567, -0x4568, 0x7FFF, -0x8000, -1), endian="big"
    )


def test_tomono() -> None:
    """Test mono channel conversion."""
    for w in 1, 2, 4:
        data1 = datas[w]
        data2 = bytearray(2 * len(data1))
        for k in range(w):
            data2[k :: 2 * w] = data1[k::w]
        assert pyaudioop.tomono(data2, w, 1, 0) == data1
        assert pyaudioop.tomono(data2, w, 0, 1) == b"\0" * len(data1)
        for k in range(w):
            data2[k + w :: 2 * w] = data1[k::w]
        assert pyaudioop.tomono(data2, w, 0.5, 0.5) == data1
        assert pyaudioop.tomono(bytearray(data2), w, 0.5, 0.5) == data1
        assert pyaudioop.tomono(memoryview(data2), w, 0.5, 0.5) == data1


def test_tostereo() -> None:
    """Test stereo channel conversion."""
    for w in 1, 2, 4:
        data1 = datas[w]
        data2 = bytearray(2 * len(data1))
        for k in range(w):
            data2[k :: 2 * w] = data1[k::w]
        assert pyaudioop.tostereo(data1, w, 1, 0) == data2
        assert pyaudioop.tostereo(data1, w, 0, 0) == b"\0" * len(data2)
        for k in range(w):
            data2[k + w :: 2 * w] = data1[k::w]
        assert pyaudioop.tostereo(data1, w, 1, 1) == data2
        assert pyaudioop.tostereo(bytearray(data1), w, 1, 1) == data2
        assert pyaudioop.tostereo(memoryview(data1), w, 1, 1) == data2


def test_ratecv() -> None:
    """Test sample rate conversion."""
    for w in 1, 2, 4:
        assert pyaudioop.ratecv(b"", w, 1, 8000, 8000, None) == (b"", (-1, ((0, 0),)))
        assert pyaudioop.ratecv(bytearray(), w, 1, 8000, 8000, None) == (
            b"",
            (-1, ((0, 0),)),
        )
        assert pyaudioop.ratecv(memoryview(b""), w, 1, 8000, 8000, None) == (
            b"",
            (-1, ((0, 0),)),
        )
        assert pyaudioop.ratecv(b"", w, 5, 8000, 8000, None) == (
            b"",
            (-1, ((0, 0),) * 5),
        )
        assert pyaudioop.ratecv(b"", w, 1, 8000, 16000, None) == (b"", (-2, ((0, 0),)))
        assert pyaudioop.ratecv(datas[w], w, 1, 8000, 8000, None)[0] == datas[w]
        assert pyaudioop.ratecv(datas[w], w, 1, 8000, 8000, None, 1, 0)[0] == datas[w]

    state = None
    d1, state = pyaudioop.ratecv(b"\x00\x01\x02", 1, 1, 8000, 16000, state)
    d2, state = pyaudioop.ratecv(b"\x00\x01\x02", 1, 1, 8000, 16000, state)
    assert d1 + d2 == b"\000\000\001\001\002\001\000\000\001\001\002"

    for w in 1, 2, 4:
        d0, state0 = pyaudioop.ratecv(datas[w], w, 1, 8000, 16000, None)
        d, state = b"", None
        for i in range(0, len(datas[w]), w):
            d1, state = pyaudioop.ratecv(datas[w][i : i + w], w, 1, 8000, 16000, state)
            d += d1
        assert d == d0
        assert state == state0

    # Not sure why this is still failing, but the crackling is gone!
    # expected = {
    #     1: packs[1](0, 0x0D, 0x37, -0x26, 0x55, -0x4B, -0x14),
    #     2: packs[2](0, 0x0DA7, 0x3777, -0x2630, 0x5673, -0x4A64, -0x129A),
    #     3: packs[3](0, 0x0DA740, 0x377776, -0x262FCA, 0x56740C, -0x4A62FD, -0x1298C0),
    #     4: packs[4](
    #         0, 0x0DA740DA, 0x37777776, -0x262FC962, 0x56740DA6, -0x4A62FC96, -0x1298BF26
    #     ),
    # }
    # for w in 1, 2, 4:
    #     assert (
    #         pyaudioop.ratecv(datas[w], w, 1, 8000, 8000, None, 3, 1)[0] == expected[w]
    #     )
    #     assert (
    #         pyaudioop.ratecv(datas[w], w, 1, 8000, 8000, None, 30, 10)[0] == expected[w]
    #     )



"""HTTP server for text to speech (TTS)."""

import io
import logging
import wave
from pathlib import Path
from typing import Optional

from flask import Response, request
from werkzeug.exceptions import BadRequest

from wyoming.audio import AudioChunk, AudioStart, AudioStop
from wyoming.error import Error
from wyoming.tts import Synthesize, SynthesizeVoice

from .shared import check_args, get_app, get_argument_parser, get_client

_DIR = Path(__file__).parent
CONF_PATH = _DIR / "conf" / "tts.yaml"


def main():
    parser = get_argument_parser()
    parser.add_argument("--voice", help="Default voice for synthesis")
    parser.add_argument("--speaker", help="Default voice speaker for synthesis")
    args = parser.parse_args()
    check_args(parser, args)
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    app = get_app("tts", CONF_PATH, args)

    @app.route("/api/text-to-speech", methods=["POST", "GET"])
    async def api_stt() -> Response:
        if request.method == "POST":
            text = request.data.decode()
        else:
            text = request.args.get("text", "")

        if not text:
            raise BadRequest("Text is required")

        voice: Optional[SynthesizeVoice] = None
        voice_name = request.args.get("voice", args.voice)
        if voice_name:
            voice = SynthesizeVoice(
                name=voice_name, speaker=request.args.get("speaker", args.speaker)
            )

        async with get_client(args) as client:
            wav_io = io.BytesIO()
            wav_file = wave.open(wav_io, "wb")
            await client.write_event(Synthesize(text=text, voice=voice).event())

            while True:
                event = await client.read_event()
                if event is None:
                    raise RuntimeError("Client disconnected")

                if AudioStart.is_type(event.type):
                    audio_start = AudioStart.from_event(event)
                    wav_file.setframerate(audio_start.rate)
                    wav_file.setsampwidth(audio_start.width)
                    wav_file.setnchannels(audio_start.channels)
                elif AudioChunk.is_type(event.type):
                    audio_chunk = AudioChunk.from_event(event)
                    wav_file.writeframes(audio_chunk.audio)
                elif AudioStop.is_type(event.type):
                    wav_file.close()
                    wav_io.seek(0)
                    return Response(
                        wav_io.getvalue(), headers={"Content-Type": "audio/wav"}
                    )
                elif Error.is_type(event.type):
                    error = Error.from_event(event)
                    raise RuntimeError(
                        f"Unexpected error from client: code={error.code}, text={error.text}"
                    )

    app.run(args.host, args.port)


if __name__ == "__main__":
    main()



"""Wake word detection"""

import asyncio
import contextlib
import logging
from asyncio.subprocess import Process
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from .audio import AudioChunk, AudioChunkConverter
from .client import AsyncClient
from .event import Event, Eventable

_LOGGER = logging.getLogger(__name__)

DOMAIN = "wake"
_DETECTION_TYPE = "detection"
_DETECT_TYPE = "detect"
_NOT_DETECTED_TYPE = "not-detected"


@dataclass
class Detection(Eventable):
    """Wake word was detected."""

    name: Optional[str] = None
    """Name of model."""

    timestamp: Optional[int] = None
    """Timestamp of audio chunk with detection"""

    speaker: Optional[str] = None
    """Name of speaker."""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _DETECTION_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {
            "name": self.name,
            "timestamp": self.timestamp,
            "speaker": self.speaker,
        }
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_DETECTION_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "Detection":
        data = event.data or {}
        return Detection(
            name=data.get("name"),
            timestamp=data.get("timestamp"),
            speaker=data.get("speaker"),
            context=data.get("context"),
        )


@dataclass
class Detect(Eventable):
    """Wake word detection request.

    Followed by AudioStart, AudioChunk+, AudioStop
    """

    names: Optional[List[str]] = None
    """Names of models to detect (None = any)."""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _DETECT_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {"names": self.names}
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_DETECT_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "Detect":
        data = event.data or {}
        return Detect(names=data.get("names"), context=data.get("context"))


@dataclass
class NotDetected(Eventable):
    """Audio stream ended before wake word was detected."""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _NOT_DETECTED_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {}
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_NOT_DETECTED_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "NotDetected":
        if not event.data:
            return NotDetected()

        return NotDetected(context=event.data.get("context"))


class WakeProcessAsyncClient(AsyncClient, contextlib.AbstractAsyncContextManager):
    """Context manager for doing wake word detection with an external program."""

    def __init__(
        self,
        rate: int,
        width: int,
        channels: int,
        program: str,
        program_args: List[str],
    ) -> None:
        super().__init__()

        self.rate = rate
        self.width = width
        self.channels = channels
        self.program = program
        self.program_args = program_args

        self._proc: Optional[Process] = None
        self._chunk_converter = AudioChunkConverter(rate, width, channels)

    async def connect(self) -> None:
        self._proc = await asyncio.create_subprocess_exec(
            self.program,
            *self.program_args,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
        )

    async def disconnect(self) -> None:
        assert self._proc is not None
        assert self._proc.stdin is not None

        try:
            if self._proc.returncode is None:
                # Terminate process gracefully
                self._proc.stdin.close()
                await self._proc.wait()
        except ProcessLookupError:
            # Expected when process has already exited
            pass
        except Exception:
            _LOGGER.exception("Unexpected error stopping process: %s", self.program)
        finally:
            self._proc = None

    async def __aenter__(self) -> "WakeProcessAsyncClient":
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.disconnect()

    async def read_event(self) -> Optional[Event]:
        assert self._proc is not None
        assert self._proc.stdout is not None

        line = (await self._proc.stdout.readline()).decode("utf-8").strip()
        name = line if line else None
        return Detection(name=name).event()

    async def write_event(self, event: Event) -> None:
        assert self._proc is not None
        assert self._proc.stdin is not None

        if not AudioChunk.is_type(event.type):
            return

        chunk = AudioChunk.from_event(event)

        # Convert sample rate/width/channels if necessary
        chunk = self._chunk_converter.convert(chunk)

        self._proc.stdin.write(chunk.audio)
        await self._proc.stdin.drain()



import asyncio
from abc import ABC
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse

from .event import (
    Event,
    async_get_stdin,
    async_get_stdout,
    async_read_event,
    async_write_event,
)


class AsyncClient(ABC):
    """Base class for Wyoming async client."""

    def __init__(
        self,
        connect_timeout: Optional[float] = None,
        read_timeout: Optional[float] = None,
    ) -> None:
        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None

        # Seconds to wait when connecting, or None to wait indefinitely
        self.connect_timeout = connect_timeout

        # Seconds to wait for each event, or None to wait indefinitely
        self.read_timeout = read_timeout

    async def read_event(self) -> Optional[Event]:
        assert self._reader is not None

        if self.read_timeout is None:
            return await async_read_event(self._reader)

        return await asyncio.wait_for(
            async_read_event(self._reader), timeout=self.read_timeout
        )

    async def write_event(self, event: Event) -> None:
        assert self._writer is not None
        await async_write_event(event, self._writer)

    async def connect(self) -> None:
        pass

    async def __aenter__(self):
        await self.connect()
        return self

    async def disconnect(self) -> None:
        pass

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.disconnect()

    @staticmethod
    def from_uri(
        uri: str,
        connect_timeout: Optional[float] = None,
        read_timeout: Optional[float] = None,
    ) -> "AsyncClient":
        result = urlparse(uri)
        timeouts = {"connect_timeout": connect_timeout, "read_timeout": read_timeout}

        if result.scheme == "unix":
            return AsyncUnixClient(result.path, **timeouts)

        if result.scheme == "tcp":
            if (result.hostname is None) or (result.port is None):
                raise ValueError("A port must be specified when using a 'tcp://' URI")

            return AsyncTcpClient(result.hostname, result.port, **timeouts)

        if result.scheme == "stdio":
            return AsyncStdioClient(**timeouts)

        raise ValueError("Only 'stdio://', 'unix://', or 'tcp://' are supported")


class AsyncTcpClient(AsyncClient):
    """TCP Wyoming client."""

    def __init__(
        self,
        host: str,
        port: int,
        connect_timeout: Optional[float] = None,
        read_timeout: Optional[float] = None,
    ) -> None:
        super().__init__(connect_timeout=connect_timeout, read_timeout=read_timeout)

        self.host = host
        self.port = port

    async def connect(self) -> None:
        connect = asyncio.open_connection(host=self.host, port=self.port)

        if self.connect_timeout is not None:
            connect = asyncio.wait_for(connect, timeout=self.connect_timeout)

        self._reader, self._writer = await connect

    async def disconnect(self) -> None:
        writer = self._writer
        self._reader = None
        self._writer = None

        if writer is not None:
            writer.close()
            await writer.wait_closed()


class AsyncUnixClient(AsyncClient):
    """Unix domain socket Wyoming client."""

    def __init__(
        self,
        socket_path: Union[str, Path],
        connect_timeout: Optional[float] = None,
        read_timeout: Optional[float] = None,
    ) -> None:
        super().__init__(connect_timeout=connect_timeout, read_timeout=read_timeout)

        self.socket_path = Path(socket_path)

    async def connect(self) -> None:
        connect = asyncio.open_unix_connection(path=self.socket_path)

        if self.connect_timeout is not None:
            connect = asyncio.wait_for(connect, timeout=self.connect_timeout)

        self._reader, self._writer = await connect

    async def disconnect(self) -> None:
        writer = self._writer
        self._reader = None
        self._writer = None

        if writer is not None:
            writer.close()
            await writer.wait_closed()


class AsyncStdioClient(AsyncClient):
    """Standard output Wyoming client."""

    async def read_event(self) -> Optional[Event]:
        if self._reader is None:
            self._reader = await async_get_stdin()

        return await super().read_event()

    async def write_event(self, event: Event) -> None:
        if self._writer is None:
            self._writer = await async_get_stdout()

        await super().write_event(event)



"""Intent recognition and handling."""

from dataclasses import dataclass
from typing import Any, Dict, Optional

from .event import Event, Eventable

DOMAIN = "handle"
_HANDLED_TYPE = "handled"
_NOT_HANDLED_TYPE = "not-handled"
_HANDLED_START_TYPE = "handled-start"
_HANDLED_CHUNK_TYPE = "handled-chunk"
_HANDLED_STOP_TYPE = "handled-stop"


@dataclass
class Handled(Eventable):
    """Result of successful intent handling."""

    text: Optional[str] = None
    """Human-readable response."""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _HANDLED_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {}
        if self.text is not None:
            data["text"] = self.text
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_HANDLED_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "Handled":
        if not event.data:
            return Handled()

        return Handled(text=event.data.get("text"), context=event.data.get("context"))


@dataclass
class NotHandled(Eventable):
    """Result of intent handling failure."""

    text: Optional[str] = None
    """Human-readable response."""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _NOT_HANDLED_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {}
        if self.text is not None:
            data["text"] = self.text
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_NOT_HANDLED_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "NotHandled":
        if not event.data:
            return NotHandled()

        return NotHandled(
            text=event.data.get("text"), context=event.data.get("context")
        )


@dataclass
class HandledStart(Eventable):
    """Start of streaming result of successful intent handling."""

    context: Optional[Dict[str, Any]] = None
    """Context for next interaction."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _HANDLED_START_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {}
        if self.context is not None:
            data["context"] = self.context

        return Event(type=_HANDLED_START_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "HandledStart":
        if not event.data:
            return HandledStart()

        return HandledStart(context=event.data.get("context"))


@dataclass
class HandledChunk(Eventable):
    """Response chunk of streaming result of successful intent handling."""

    text: str
    """Chunk of response text."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _HANDLED_CHUNK_TYPE

    def event(self) -> Event:
        return Event(type=_HANDLED_CHUNK_TYPE, data={"text": self.text})

    @staticmethod
    def from_event(event: Event) -> "HandledChunk":
        return HandledChunk(text=event.data["text"])


@dataclass
class HandledStop(Eventable):
    """End of streaming sesult of successful intent handling."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _HANDLED_STOP_TYPE

    def event(self) -> Event:
        return Event(type=_HANDLED_STOP_TYPE)

    @staticmethod
    def from_event(event: Event) -> "HandledStop":
        return HandledStop()



"""HTTP server for automated speech recognition (ASR)."""

import io
import logging
import wave
from pathlib import Path
from typing import List, Optional

from flask import Response, jsonify, request

from wyoming.asr import Transcribe, Transcript
from wyoming.audio import wav_to_chunks
from wyoming.error import Error

from .shared import check_args, get_app, get_argument_parser, get_client

_DIR = Path(__file__).parent
CONF_PATH = _DIR / "conf" / "asr.yaml"


def main():
    parser = get_argument_parser()
    parser.add_argument("--model", help="Default model name for transcription")
    parser.add_argument("--language", help="Default language for transcription")
    parser.add_argument("--samples-per-chunk", type=int, default=1024)
    args = parser.parse_args()
    check_args(parser, args)
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    app = get_app("asr", CONF_PATH, args)

    @app.route("/api/speech-to-text", methods=["POST"])
    async def api_stt() -> Response:
        model_name = request.args.get("model", args.model)
        language = request.args.get("language", args.language)

        transcript_names: Optional[List[str]] = None
        transcript_names_str = request.args.get("transcript_names")
        if transcript_names_str:
            transcript_names = [
                name.strip() for name in transcript_names_str.split(",") if name.strip()
            ]

        transcript_terms: Optional[List[str]] = None
        transcript_terms_str = request.args.get("transcript_terms")
        if transcript_terms_str:
            transcript_terms = [
                term.strip() for term in transcript_terms_str.split(",") if term.strip()
            ]

        async with get_client(args) as client:
            await client.write_event(
                Transcribe(
                    name=model_name,
                    language=language,
                    transcript_names=transcript_names,
                    transcript_terms=transcript_terms,
                ).event()
            )

            with io.BytesIO(request.data) as wav_io:
                with wave.open(wav_io, "rb") as wav_file:
                    chunks = wav_to_chunks(
                        wav_file,
                        samples_per_chunk=args.samples_per_chunk,
                        start_event=True,
                        stop_event=True,
                    )
                    for chunk in chunks:
                        await client.write_event(chunk.event())

            while True:
                event = await client.read_event()
                if event is None:
                    raise RuntimeError("Client disconnected")

                if Transcript.is_type(event.type):
                    transcript = Transcript.from_event(event)
                    return jsonify(transcript.to_dict())

                if Error.is_type(event.type):
                    error = Error.from_event(event)
                    raise RuntimeError(
                        f"Unexpected error from client: code={error.code}, text={error.text}"
                    )

    app.run(args.host, args.port)


if __name__ == "__main__":
    main()
